#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import logging.config
import os
import re
import sys
import threading
import traceback

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

#---------------------------------------
# Docker-exec runner
#---------------------------------------
class ExecRunner(threading.Thread):

    def __init__(self, url, cid, command):
        super(ExecRunner, self).__init__()
        self.__url = url
        self.__cid = cid
        self.__command = command
        self.container_id = cid # 外部からの参照用

    def run(self):
        dc = docker.Client(base_url=self.__url, version=DOCKER_API_VERSION)

        # exec create
        exec_info = dc.exec_create(
                container=self.__cid, cmd=self.__command,
                stdout=True, stderr=True, tty=False)
        logger.debug("exec_info:%s", exec_info)

        # exec start
        for output in dc.exec_start(
                exec_id=exec_info["Id"], detach=False, tty=False,
                stream=True):   # stream=Falseだとタイムアウトする可能性がある
            sys.stdout.write(output.decode("utf-8"))  # printだと余分な改行が入る
            sys.stdout.flush()

        # exec inspect
        insp = dc.exec_inspect(exec_id=exec_info["Id"])
        logger.debug("exec_inspect:%s", insp)

        self.result = insp["ExitCode"]

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerコンテナでコマンドを実行します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
            "-c", dest="cid", metavar="CONTAINER_ID", 
            help="コンテナのIDパターンを指定します。")
    group.add_argument(
            "-n", dest="name", metavar="CONTAINER_NAME",
            help="コンテナの名称パターンを指定します。")
    parser.add_argument(
            "-u", dest="user", metavar="USER",
            help="コマンドを実行するユーザを指定します。")
    parser.add_argument(
            "command", metavar="COMMAND",
            help="実行するコマンドを指定します。")

    args = parser.parse_args()

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # コマンドを加工する(ユーザ指定ありの場合はsudo経由)
    if args.user is None:
        args.command2 = ["sh", "-c", args.command]
    else:
        args.command2 = ["sudo", "-u", args.user, "sh", "-c", args.command]

    # 引数を記録
    logger.info("args:%s", args)

    # コンテナIDのリストを作成する
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    containers = []
    if args.cid :
        pa = re.compile(('%s.*' % args.cid[:len(args.cid) - 1]) if args.cid.endswith('*') else args.cid)
        def find_container_by_id(c):
            if pa.match(c["Id"]):
                containers.append(c["Id"])
        map(find_container_by_id, dc.containers(quiet=True))
    elif args.name :
        pa = re.compile(('%s.*' % args.name[:len(args.name) - 1]) if args.name.endswith('*') else args.name)
        def find_container_by_name(c):
            # 名前に含まれる '/' を除外するため、マッチングする文字列の 1 文字目をスキップ
            matched_names = [n[1:] for n in c["Names"] if pa.match(n[1:])]
            if len(matched_names) > 0:
                containers.append(c["Id"])
        map(find_container_by_name, dc.containers())
    else :
        def all_container(c):
            containers.append(c["Id"])
        map(all_container,dc.containers(quiet=True))
        
    if len(containers) == 0:
        raise LocalException("指定されたコンテナは存在しません。")

    # コマンド実行
    runners = []
    for cid in containers:
        runners.append(ExecRunner(args.url, cid, args.command2))

    for r in runners: r.start()
    for r in runners: r.join()

    return runners[0].result    #暫定

#===============================================================================
if __name__ == "__main__":
    try:
        exit_code = main()
        sys.exit(exit_code)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
