#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import logging.config
import os
import sys
import threading
import traceback
import re

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

#---------------------------------------
# Docker-exec runner
#---------------------------------------
class ExecRunner(threading.Thread):

    def __init__(self, url, cid, command):
        super(ExecRunner, self).__init__()
        self.__url = url
        self.__cid = cid
        self.__command = command
        self.container_id = cid # 外部からの参照用

    def run(self):
        dc = docker.Client(base_url=self.__url, version=DOCKER_API_VERSION)

        # exec create
        exec_info = dc.exec_create(
                container=self.__cid, cmd=self.__command,
                stdout=True, stderr=True, tty=False)
        logger.debug("exec_info:%s", exec_info)

        # exec start
        for output in dc.exec_start(
                exec_id=exec_info["Id"], detach=False, tty=False,
                stream=True):   # stream=Falseだとタイムアウトする可能性がある
            sys.stdout.write(output.decode("utf-8"))  # printだと余分な改行が入る
            sys.stdout.flush()

        # exec inspect
        insp = dc.exec_inspect(exec_id=exec_info["Id"])
        logger.debug("exec_inspect:%s", insp)

        self.result = insp["ExitCode"]

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerコンテナでコマンドを実行します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
            "-c", dest="cid", metavar="CONTAINER_ID", 
            help="コンテナのIDパターンを指定します。")
    group.add_argument(
            "-n", dest="name", metavar="CONTAINER_NAME",
            help="コンテナの名称パターンを指定します。")
    
    group_f = parser.add_mutually_exclusive_group()
    group_f.add_argument(
            "-f", dest="force", action='store_true',
            help="起動中のコンテナも強制的に削除します。")
    group_f.add_argument(
            "--force=false", dest="force", action='store_false',
            help="停止しているコンテナを削除します。（デフォルト動作）")

    group_l = parser.add_mutually_exclusive_group()
    group_l.add_argument(
            "-l", dest="link", action='store_true',
            help="リンク先のコンテナを削除します。")
    group_l.add_argument(
            "--link=false", dest="link", action='store_false',
            help="リンク先を削除しません。（デフォルト動作）")
    
    group_v = parser.add_mutually_exclusive_group()
    group_v.add_argument(
            "-v", dest="volumes", action='store_true',
            help="コンテナに関連付けられているボリュームも削除します。")
    group_v.add_argument(
            "--volumes=false", dest="volumes", action='store_false',
            help="コンテナに関連するボリュームを削除しません。（デフォルト動作）")

    args = parser.parse_args()

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # 引数を記録
    logger.info("args:%s", args)

    return rm_by_id(args.url, args.cid, args.force, args.link, args.volumes) if args.cid else rm_by_name(args.url, args.name, args.force, args.link, args.volumes)


def rm_by_id(docker_url, id_pattern, f, l, v):
    dc = docker.Client(base_url=docker_url, version=DOCKER_API_VERSION)
    pa = re.compile(('%s.*' % id_pattern[:len(id_pattern) - 1]) if id_pattern.endswith('*') else id_pattern)
    
    def rm_container_by_id(c):
      if (not pa.match(c['Id'])):
          return False
      
      try:
          dc.remove_container(c['Id'], v=v, link=l, force=f)
          print '%s %s %s' % (c['Id'][:12], c['Names'][0][1:], 'Success')
          return True
      except Exception, e:
          print '%s %s %s' % (c['Id'][:12], c['Names'][0][1:], 'Fail')
          print e
          return False

    r = map(rm_container_by_id, dc.containers(all=True))
    code = 0 if r.count(True) == len(r) else (9 if r.count(False) == len(r) else 1)
    #logger.info("exsit code:%s", code)
    return code
  
  
def rm_by_name(docker_url, name_pattern, f, l, v):
    dc = docker.Client(base_url=docker_url, version=DOCKER_API_VERSION)
    pa = re.compile(('%s.*' % name_pattern[:len(name_pattern) - 1]) if name_pattern.endswith('*') else name_pattern)
    
    def rm_container_by_name(c):
      # 名前に含まれる '/' を除外するため、マッチングする文字列の 1 文字目をスキップ
      matched_names = [n[1:] for n in c['Names'] if pa.match(n[1:])]
      if (len(matched_names) == 0):
          return False
      
      try:
          dc.remove_container(c['Id'], v=v, link=l, force=f)
          print '%s %s %s' % (c['Id'][:12], matched_names[0], 'Success')
          return True
      except Exception, e:
          print '%s %s %s' % (c['Id'][:12], matched_names[0], 'Fail')
          print e
          return False

    r = map(rm_container_by_name, dc.containers(all=True))
    code = 0 if r.count(True) == len(r) else (9 if r.count(False) == len(r) else 1)
    #logger.info("exsit code:%s", code)
    return code


#===============================================================================
if __name__ == "__main__":
    try:
        sys.exit(main())

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)