#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import json
import logging.config
import netsnmp
import os
import re
import sys
import threading
import traceback

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
DEFAULT_SNMP_VER = 1
DEFAULT_SNMP_COM = "public"
DEFAULT_SNMP_PORT = "161"
CID_LEN = 8

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

#---------------------------------------
# Docker-stats poller
#---------------------------------------
class StatsPoller(threading.Thread):

    def __init__(self, url, cid):
        super(StatsPoller, self).__init__()
        self.__url = url
        self.__cid = cid

    def run(self):
        c = docker.Client(base_url=self.__url, version=DOCKER_API_VERSION)
        for stats_json in c.stats(self.__cid):
            logger.debug("StatsPoller; stats:%s", stats_json.rstrip())
            self.result = json.loads(stats_json);
            break

#---------------------------------------
# SNMP poller
#---------------------------------------
class SnmpPoller(threading.Thread):

    def __init__(self, host, ver, com, port, *oidlist):
        super(SnmpPoller, self).__init__()
        self.__host = host
        self.__ver = ver
        self.__com = com
        self.__port = port
        self.__oidlist = oidlist

    def run(self):
        res = netsnmp.snmpget(
                *self.__oidlist,
                Version=self.__ver,
                DestHost="%s:%s" % (self.__host, self.__port), # RemotePortパラメータは無視される？
                Community=self.__com)
        self.result = res
        logger.debug("SnmpPoller; host:%s ver:%s com:%s port:%s oids:%s result:%s",
                     self.__host, self.__ver, self.__com, self.__port, str(self.__oidlist), str(res))

    def get_int(self, index):
        r = self.result[index]
        if r is None or r == "":
            raise LocalException("SNMPで値が取得できませんでした。OID:%s" % (self.__oidlist[index]))
        return int(r)

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerのメモリ情報を取得します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    parser.add_argument(
            "-sv", dest="snmp_ver", metavar="SNMP_VER", type=int,
            default=DEFAULT_SNMP_VER,
            help="SNMPのバージョン。(default: '%s')" % DEFAULT_SNMP_VER)
    parser.add_argument(
            "-sc", dest="snmp_com", metavar="SNMP_COM",
            default=DEFAULT_SNMP_COM,
            help="SNMPのコミュニティ。(default: '%s')" % DEFAULT_SNMP_COM)
    parser.add_argument(
            "-sp", dest="snmp_port", metavar="SNMP_PORT",
            default=DEFAULT_SNMP_PORT,
            help="SNMPのポート。(default: '%s')" % DEFAULT_SNMP_PORT)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
            "-c", dest="cid", metavar="CONTAINER_ID", 
            help="コンテナのIDパターンを指定します。")
    group.add_argument(
            "-n", dest="name", metavar="CONTAINER_NAME",
            help="コンテナの名称パターンを指定します。")
    parser.add_argument(
            "--container", dest="container_limit", action="store_true",
            help="コンテナの設定上限に対するメモリ使用率を出力します。")
    parser.add_argument(
            "--with-host", dest="with_host", action="store_true",
            help="コンテナ以外のホストのメモリ使用率を合わせて出力します。")
    parser.add_argument(
            "--all", dest="detail_info", action="store_true",
            help="コンテナのメモリ関連詳細情報を出力します。")

    args = parser.parse_args()

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # --with-host は、-c と -n と --container 不使用時のみ
    if args.with_host and (args.cid or args.name or args.container_limit):
        raise LocalException("--with-hostは-c,-n,--containerと同時指定できません。")

    # コンテナIDのリストを作成する
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    containers = []
    if args.cid :
        pa = re.compile(('%s.*' % args.cid[:len(args.cid) - 1]) if args.cid.endswith('*') else args.cid)
        def find_container_by_id(c):
            if pa.match(c["Id"]):
                containers.append(c["Id"])
        map(find_container_by_id, dc.containers(quiet=True))
    elif args.name :
        pa = re.compile(('%s.*' % args.name[:len(args.name) - 1]) if args.name.endswith('*') else args.name)
        def find_container_by_name(c):
            # 名前に含まれる '/' を除外するため、マッチングする文字列の 1 文字目をスキップ
            matched_names = [n[1:] for n in c["Names"] if pa.match(n[1:])]
            if len(matched_names) > 0:
                containers.append(c["Id"])
        map(find_container_by_name, dc.containers())
    else :
        def all_container(c):
            containers.append(c["Id"])
        map(all_container,dc.containers(quiet=True))
        
    if len(containers) == 0:
        raise LocalException("指定されたコンテナは存在しません。")

    # --all 指定時は、単体コンテナのみ指定
    if args.detail_info and len(containers) != 1:
        raise LocalException("--all指定時はコンテナを1つだけ指定してください。")

    # 引数を記録
    logger.info("args:%s", args)

    # 処理分岐
    if args.detail_info:
        main_detail_info(args, containers)
    else:
        main_memory_usage(args, containers)

#---------------------------------------
# メイン：メモリ詳細情報
#---------------------------------------
def main_detail_info(args, containers):

    # ポーリング実行
    poller = StatsPoller(args.url, containers[0])
    poller.start()
    poller.join()

    # 結果出力
    stats = poller.result["memory_stats"]["stats"]
    for key in sorted(stats.iterkeys()):
        print "%s,%s" % (key, stats[key])

#---------------------------------------
# メイン：メモリ使用率
#---------------------------------------
def main_memory_usage(args, containers):

    # コンテナ上限を基準とする場合は、inspectのため、Dockerクライアントを準備する
    if args.container_limit:
        dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)

    # ポーリング実行
    # pollers: docker-stats poller(複数あり)
    # snmp   : SNMP poller
    pollers = []
    for cid in containers:
        pollers.append(StatsPoller(args.url, cid))

    snmp = SnmpPoller(args.ip_addr, args.snmp_ver, args.snmp_com, args.snmp_port,
                      ".1.3.6.1.4.1.2021.4.5.0",  # memTotalReal
                      ".1.3.6.1.4.1.2021.4.6.0",  # memAvailReal
                      ".1.3.6.1.4.1.2021.4.14.0", # memBuffer
                      ".1.3.6.1.4.1.2021.4.15.0") # memCached

    for p in pollers: p.start()
    snmp.start()

    for p in pollers: p.join()
    snmp.join()

    # SNMPの結果を取得する (ついでに単位変換 KB → B)
    snmp_memTotalReal = snmp.get_int(0) * 1024
    snmp_memAvailReal = snmp.get_int(1) * 1024
    snmp_memBuffer    = snmp.get_int(2) * 1024
    snmp_memCached    = snmp.get_int(3) * 1024

    # コンテナごとの値を出力
    sum_container_rss = 0
    for cid, p in zip(containers, pollers):

        # メモリ上限を求める
        if args.container_limit:
            insp = dc.inspect_container(cid)
            limit = insp["Config"]["Memory"]
            # 上限指定なしであれば、ホストのメモリ量を上限とする
            if limit == 0: limit = snmp_memTotalReal
        else:
            limit = snmp_memTotalReal

        # 計算＆出力
        rss = p.result["memory_stats"]["stats"]["rss"]
        logger.debug("cid:%s, limit:%s, rss:%s", cid, limit, rss)
        sum_container_rss += rss
        memused = float(rss) / limit * 100
        if args.cid or args.name is not None:
            # コンテナ指定の場合はkeyを出力しない
            print "host_mem_usage,%.2f" % (memused)
        else:
            print "host_mem_usage:%s,%.2f" % (cid[:CID_LEN], memused)

    # ホスト値を出力
    if args.with_host:
        memused = float(snmp_memTotalReal - snmp_memAvailReal - snmp_memBuffer - snmp_memCached
                        - sum_container_rss) / snmp_memTotalReal * 100
        print "host_mem_usage:host,%.2f" % (memused)

#===============================================================================
if __name__ == "__main__":
    try:
        main()
        sys.exit(0)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
