#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import logging.config
import os
import pprint
import suds
import sys
import syslog
import traceback

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
DEFAULT_HINEMOS_URL = "http://127.0.0.1:8080/HinemosWS"
CID_LEN = 12

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

# 開発/デバッグ用
def dumpvar(var):
    print pprint.pformat(var).decode("utf-8")

#---------------------------------------
# メッセージ出力
#---------------------------------------
def trace(message, to_syslog=False):
    # syslog
    if to_syslog: syslog.syslog(syslog.LOG_INFO, message.encode("utf-8"))
    # log
    logger.debug(message)
    # console
    print message

#---------------------------------------
# mapスコープ配下のスコープ名を返す
#---------------------------------------
def get_map_name(host_fid):
    return "%s_scope" % host_fid

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerノードを自動削除します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    parser.add_argument(
            "-m", dest="hinemos_url", metavar="HINEMOS_URL",
            default=DEFAULT_HINEMOS_URL,
            help="HinemosマネージャのURL。(default: '%s')" % DEFAULT_HINEMOS_URL)
    parser.add_argument(
            "-f", dest="facility_id", metavar="HOST_FACILITY_ID",
            help="Dockerホストの対応するノードのファシリティID。")

    args = parser.parse_args()

    # ファシリティIDは必須
    if args.facility_id is None:
        raise LocalException("ファシリティIDの指定(-f)は必須です。")

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # Hinemos URLの末尾に"/"がある場合は除去
    if args.hinemos_url.endswith("/"):
        args.hinemos_url = args.hinemos_url[:-1]

    # Hinemosへのログイン情報
    args.hinemos_username = "hinemos"
    args.hinemos_password = "hinemos"

    # 引数を記録
    logger.info("args:%s", args)

    # 主処理
    syslog.openlog(str("RepositoryDockerNodeDelete"))
    trace("開始しました。", True)
    try:
        hclient = login_hinemos(args.hinemos_url, args.hinemos_username, args.hinemos_password)

        dclient = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
        containers = get_containers_info(dclient, args.facility_id)

        delete_container_nodes(hclient, args.facility_id, containers)

        trace("完了しました。", True)

    except:
        exc_info = sys.exc_info()
        trace("Error: %s,%s" % (exc_info[0], exc_info[1]), True)
        raise

    finally:
        syslog.closelog()

#---------------------------------------
# Hinemosマネージャへログイン
#---------------------------------------
def login_hinemos(url, username, password):
    trace("Hinemosマネージャにログインしています。")

    endpoint = "%s/RepositoryEndpoint?wsdl" % url

    trace("  endpoint: %s" % endpoint)
    trace("  username: %s" % username)

    hclient = suds.client.Client(
            endpoint,
            transport=suds.transport.http.HttpAuthenticated(
                username=username, password=password))

    return hclient

#---------------------------------------
# Docker コンテナ情報を取得する
#---------------------------------------
def get_containers_info(dclient, host_fid):
    trace("Dockerコンテナの情報を取得します。")

    containers = dclient.containers(all=True)

    # 後から参照する値をここで算出し、dict へ追加設定する(prefix = "+")
    for it in containers:

        # ファシリティID
        it["+facility_id"] = "%s-%s" % (host_fid, it["Names"][0][1:])

        trace("  Id:%s Name:%s" % (it["Id"][:CID_LEN], it["Names"][0][1:]))

    return containers

#---------------------------------------
# Dockerコンテナノードを削除する
#---------------------------------------
def delete_container_nodes(hclient, host_fid, containers):
    trace("Dockerコンテナノードを削除します。")

    # コンテナのファシリティIDのリストを作成
    container_fids = []
    for container in containers:
        container_fids.append(container["+facility_id"])

    # mapスコープからコンテナノードのファシリティIDを取得する
    map_scope = get_map_name(host_fid)
    node_fids = hclient.service.getFacilityIdList(map_scope, 1)
    node_fids.remove(map_scope)

    # コンテナが存在しないノードを削除する
    for node_fid in node_fids:
        trace("  %s" % node_fid)
        if node_fid in container_fids:
            trace("    存在します。[skip]")
        else:
            trace("    削除します。")
            hclient.service.deleteNode(node_fid)

#===============================================================================
if __name__ == "__main__":
    try:
        exit_code = main()
        sys.exit(exit_code)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
