#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import logging.config
import os
import pprint
import suds
import sys
import syslog
import traceback

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
DEFAULT_HINEMOS_URL = "http://127.0.0.1:8080/HinemosWS"
CID_LEN = 12

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

# 開発/デバッグ用
def dumpvar(var):
    print pprint.pformat(var).decode("utf-8")

#---------------------------------------
# メッセージ出力
#---------------------------------------
def trace(message, to_syslog=False):
    # syslog
    if to_syslog: syslog.syslog(syslog.LOG_INFO, message.encode("utf-8"))
    # log
    logger.debug(message)
    # console
    print message

#---------------------------------------
# Dockerコンテナのステータスを、human-readableな表現から単語表現へ変換する
#---------------------------------------
def conv_docker_status(status):
    # 起動エラーによりステータスが空欄となることがある
    if status == "": return "exited"
    w = status.split(" ")
    if w[0] == "Exited": return "exited"
    if w[0] == "Restarting": return "restarting"
    if w[0] == "Up":
        if w[-1] == "(Paused)": return "paused"
        return "running"
    raise LocalException("変換できないステータス:%s" % status)

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerノードを自動更新します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    parser.add_argument(
            "-m", dest="hinemos_url", metavar="HINEMOS_URL",
            default=DEFAULT_HINEMOS_URL,
            help="HinemosマネージャのURL。(default: '%s')" % DEFAULT_HINEMOS_URL)
    parser.add_argument(
            "-f", dest="facility_id", metavar="HOST_FACILITY_ID",
            help="Dockerホストの対応するノードのファシリティID。")

    args = parser.parse_args()

    # ファシリティIDは必須
    if args.facility_id is None:
        raise LocalException("ファシリティIDの指定(-f)は必須です。")

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # Hinemos URLの末尾に"/"がある場合は除去
    if args.hinemos_url.endswith("/"):
        args.hinemos_url = args.hinemos_url[:-1]

    # Hinemosへのログイン情報
    args.hinemos_username = "hinemos"
    args.hinemos_password = "hinemos"

    # 引数を記録
    logger.info("args:%s", args)

    # 主処理
    syslog.openlog(str("RepositoryDockerNodeUpdate"))
    trace("開始しました。", True)
    try:
        
        hclient = login_hinemos(args.hinemos_url, args.hinemos_username, args.hinemos_password)

        
        dclient = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
        containers = get_containers_info(dclient, args.facility_id)

        update_container_nodes(hclient, args.facility_id, containers)
        update_container_scope(hclient, containers)
        trace("完了しました。", True)

    except:
        exc_info = sys.exc_info()
        trace("Error: %s,%s" % (exc_info[0], exc_info[1]), True)
        raise

    finally:
        syslog.closelog()

#---------------------------------------
# Hinemosマネージャへログイン
#---------------------------------------
def login_hinemos(url, username, password):
    trace("Hinemosマネージャにログインしています。")

    endpoint = "%s/RepositoryEndpoint?wsdl" % url

    trace("  endpoint: %s" % endpoint)
    trace("  username: %s" % username)

    hclient = suds.client.Client(
            endpoint,
            transport=suds.transport.http.HttpAuthenticated(
                username=username, password=password))

    return hclient

#---------------------------------------
# Docker コンテナ情報を取得する
#---------------------------------------
def get_containers_info(dclient, host_id):
    trace("Dockerコンテナの情報を取得します。")

    containers = dclient.containers(all=True)

    # 後から参照する値をここで算出し、dict へ追加設定する(prefix = "+")
    for it in containers:

        # inspect情報
#        it["+inspect"] = dclient.inspect_container(it["Id"])

        # ファシリティID
        it["+facility_id"] = "%s-%s" % (host_id, it["Names"][0][1:])

        # ステータス名
        it["+status_name"] = conv_docker_status(it["Status"])

        trace("  Id:%s Name:%s Status:'%s':%s" % (
                it["Id"][:CID_LEN], it["Names"][0][1:], it["Status"], it["+status_name"]))

    return containers

#---------------------------------------
# Dockerコンテナノードを更新する
#---------------------------------------
def update_container_nodes(hclient, facility_id, containers):
    trace("Dockerコンテナノードを更新します。")

    for container in containers:
        trace("  %s" % container["+facility_id"])

        # ノード情報を取得
        try:
            node_info = hclient.service.getNode(container["+facility_id"])
        except suds.WebFault as e:
            # エラーメッセージが存在しないファシリティIDを示す内容の場合は
            # 無視して次のコンテナを処理する
            if "FacilityEntity.findByPrimaryKey" in str(e):
                trace("    ノードが存在しません")
                continue
            raise

        # ノード変数"DOCKER_STATUS"を比較し、更新の必要があるかを判定する
        skips = False
        target_nodevar = None

        for it in node_info.nodeVariableInfo:
            if it.nodeVariableName == "DOCKER_STATUS":
                target_nodevar = it
                if it.nodeVariableValue == container["+status_name"]:
                    trace("    %s -> %s [skip]" % (it.nodeVariableValue, container["+status_name"]))
                    skips = True
                    break
        if skips: continue

        if target_nodevar is None:
            # DOCKER_STATUS が存在していない場合は新規に作成する
            target_nodevar = hclient.factory.create("nodeVariableInfo")
            target_nodevar.nodeVariableName = "DOCKER_STATUS"
            node_info.nodeVariableInfo.append(target_nodevar)

        # ノード情報を更新する
        trace("    %s -> %s" % (target_nodevar.nodeVariableValue, container["+status_name"]))

        target_nodevar.nodeVariableValue = container["+status_name"]
        node_info.valid = (container["+status_name"] == "running")
        hclient.service.modifyNode(node_info)
        
#---------------------------------------
# Dockerコンテナの所属スコープを更新する
#---------------------------------------
def update_container_scope(hclient, containers):
    exit_scope_ids = hclient.service.getFacilityIdList(None, 1)
    idlist={}
    for exit_scope_id in exit_scope_ids :
        idlist[exit_scope_id]=hclient.service.getFacilityIdList(exit_scope_id, 1)

    for container in containers :
        facility_id = container["+facility_id"]
        scope_ids = container["Labels"]["hinemosAssignScopeId"].split(",") if container["Labels"].has_key("hinemosAssignScopeId") else []

        if not facility_id in hclient.service.getFacilityIdList(facility_id, 1) :
            trace("[skip] %s が存在しません" % facility_id)
            continue

        trace("%s のラベル設定の確認 hinemosAssignScopeId: %s" % (facility_id, scope_ids))
        for label_id in scope_ids :
            if label_id in exit_scope_ids :
                if facility_id in idlist[label_id] :
                    trace("  [skip] 既にスコープ %s に割当てられています" % label_id)
                else :
                    trace("  %s スコープに %s を割当てます" % (label_id, facility_id))
                    hclient.service.assignNodeScope(label_id, [facility_id])
            else :
                trace("  [skip] スコープ %s が存在していないため、割当てをスキップします" % label_id)

#===============================================================================
if __name__ == "__main__":
    try:
        exit_code = main()
        sys.exit(exit_code)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
